module Queue where

import qualified SpecQueue as Spec
import Test.QuickCheck

------------------------------------------------------------------

empty   :: Q a
add     :: a -> Q a -> Q a
remove  :: Q a -> Q a
front   :: Q a -> a
isEmpty :: Q a -> Bool

------------------------------------------------------------------

data Q a = Q [a] [a]
  deriving (Eq, Show)
  
empty                     = Q [] []
--add x (Q front back)      = fixQ front (x:back)
add x (Q front back)      = Q front (x:back)
remove (Q (x:front) back) = fixQ front back
front (Q (x:front) back)  = x
isEmpty (Q front back)    = null front && null back
  
------------------------------------------------------------------

fixQ :: [a] -> [a] -> Q a
fixQ []    back = Q (reverse back) []
fixQ front back = Q front back

------------------------------------------------------------------

instance Arbitrary a => Arbitrary (Q a) where
  arbitrary =
    do front <- arbitrary
       back  <- arbitrary
       return (Q front back) -- (if null front then [] else back))

------------------------------------------------------------------

contents :: Q Int -> Spec.Q Int
contents (Q front back) = Spec.Q (front ++ reverse back)

invariant :: Q Int -> Bool
invariant (Q front back) = not (null front && not (null back))

------------------------------------------------------------------

prop_Invariant q =
  invariant q

------------------------------------------------------------------

prop_Empty =
  contents empty == Spec.empty

prop_Add x q =
  contents (add x q) == Spec.add x (contents q)

prop_Remove q =
  not (isEmpty q) ==>
    contents (remove q) == Spec.remove (contents q)
      
prop_Front q = 
  not (isEmpty q) ==>
    front q == Spec.front (contents q)

prop_IsEmpty q =
  isEmpty q == Spec.isEmpty (contents q)

------------------------------------------------------------------

prop_Empty_Inv =
  invariant empty

prop_Add_Inv x q =
  invariant (add x q)

prop_Remove_Inv q =
  not (isEmpty q) ==> 
    invariant (remove q)

------------------------------------------------------------------

